/* =============================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu
  
Example of using C language and PIC micro-controllers to solve logic equations or
design combinational blocks. A simple way to introduce the basics of low level 
programming and code organisation. 

Combinational circuit to be used to learn about how to set the
digital input and outputs of the microcontroller: Adder_BCD_1digit

============================================================================= */

// This general header file includes the device specific headers like
// "pic18f4520.h". This file contains all the microcontroller declarations.
#include <xc.h>   

#include <stdio.h>
#include "lcd.h"   // Functions to interface the LCD  

/* ------------------------------------
	Definitions 
   -------------------------------------------------------------------------- */

/* ------------------------------------
	Function prototypes 
   -------------------------------------------------------------------------- */
void init_system (void);
void read_inputs (void);
void truth_table (void);
void write_outputs (void);

/* -----------------------------------
	Global variables 
   -------------------------------------------------------------------------- */
/* Let's use convenient byte-type (char) variables. Each variable is an 8-bit
 * RAM memory position. */

static char Var_Cin;
 
static char Var_A;
static char Var_B;
static char Var_Error;

static char Var_Cin_pre;
static char Var_A_pre; // to skip the truth table calculation and printing
static char Var_B_pre; // when the same operands
static char Var_Error_pre;

static char Var_Cout;
static char Var_BCD_Units;


static char Value_data; // An 8-bit buffer where to store partial results
static char Value_data2; // An 8-bit buffer where to store partial results
static char Value_data3; // An 8-bit buffer where to store partial results


// This time to speed up the loop, we'll use this variable:
static char Var_New_Operands; // When '1' calculate and write the LCD

static char Var_LCD_flag; 	// To refresh the LCD only when there is 
							// new information to write
							
							
static int Var_Decimal; // Result translated from BCD to decimal 

// ONLY 3 bytes are necessary for printing dynamic data from 0 to 19
// for example Var_Decimal = 15 : 
//              ---> Var_Decimal_ASCII[0] = 1; 
//              ---> Var_Decimal_ASCII[1] = 5; 
//              ---> Var_Decimal_ASCII[2] = '\0'; End of string character
static char Var_Decimal_ASCII[3]; // Result translated to ASCII for printing



/* =============================================================================
		Main function 
   ========================================================================== */
void main(void){    // Let's make main() the simplest possible.
    init_system();
    while(1) {		/* loop forever reading inputs and calculating outputs*/
		read_inputs();
        truth_table();
        write_outputs();
    }		
}	

/* ------------------------------------
	Function definitions 
   -------------------------------------------------------------------------- */      

//*****************************************************************************
// Initialise the microcontroller system  
//*****************************************************************************
void init_system (void){
// Configuration bits

// These are the configuration bits translated to the XC8 Compiler
// Microcontroller general configuration bits. You don't have to change them.
// PIC18F4520 Configuration Bit Settings. They can be generated automatically 
// in the MPLAB X IDE

// CONFIG1H
#pragma config OSC = XT       // Oscillator Section bits (XT oscillator)
#pragma config FCMEN = OFF     
// Fail-Safe Clock Monitor E bit (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF      
// Internal/External Osc Switch-over bit (Oscillator Switch-over mode disabled)

// CONFIG2L
#pragma config PWRT = ON       // Power-up Timer E bit (PWRT enabled)
#pragma config BOREN = SBORDIS 
// Brown-out Reset E bits (Brown-out Reset enabled in hardware 
// only (SBOREN is disabled))
#pragma config BORV = 3    // Brown Out Reset Voltage bits (Minimum setting)

// CONFIG2H
#pragma config WDT = OFF
// Watchdog Timer E bit (WDT disabled (control is on the SWDTEN bit))
#pragma config WDTPS = 32768    
// Watchdog Timer Postscale Sect bits (1:32768)

// CONFIG3H
#pragma config CCP2MX = PORTC  
// CCP2 MUX bit (CCP2 input/output is multiplexed with RC1)
#pragma config PBADEN = OFF    
// PORTB A/D E bit (PORTB<4:0> pins are configured 
// as digital I/O on Reset)
#pragma config LPT1OSC = OFF   
// Low-Power Timer1 Oscillator E bit (Timer1 configured for 
// higher power operation)
#pragma config MCLRE = ON
// MCLR Pin E bit (MCLR pin enabled; RE3 input pin disabled)
 
// CONFIG4L
#pragma config STVREN = ON      
// Stack Full/Underflow Reset E bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF
// Single-Supply ICSP E bit (Single-Supply ICSP disabled)
#pragma config XINST = OFF
// Extended Instruction Set E bit (Instruction set extension and Indexed 
// Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF
// Code Protection bit (Block 0 (000800-001FFFh) not code-protected)
#pragma config CP1 = OFF        
// Code Protection bit (Block 1 (002000-003FFFh) not code-protected)
#pragma config CP2 = OFF        
// Code Protection bit (Block 2 (004000-005FFFh) not code-protected)
#pragma config CP3 = OFF        
// Code Protection bit (Block 3 (006000-007FFFh) not code-protected)

// CONFIG5H
#pragma config CPB = OFF        
// Boot Block Code Protection bit 
// (Boot block (000000-0007FFh) not code-protected)
#pragma config CPD = OFF        
// Data EEPROM Code Protection bit (Data EEPROM not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       
// Write Protection bit (Block 0 (000800-001FFFh) not write-protected)
#pragma config WRT1 = OFF       
// Write Protection bit (Block 1 (002000-003FFFh) not write-protected)
#pragma config WRT2 = OFF       
// Write Protection bit (Block 2 (004000-005FFFh) not write-protected)

#pragma config WRT3 = OFF       
// Write Protection bit (Block 3 (006000-007FFFh) not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       
// Configuration Register Write Protection bit (Configuration registers 
// (300000-3000FFh) not write-protected)
#pragma config WRTB = OFF
// Boot Block Write Protection bit (Boot block (000000-0007FFh) 
// not write-protected)
#pragma config WRTD = OFF
// Data EEPROM Write Protection bit (Data EEPROM not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      
// Table Read Protection bit (Block 0 (000800-001FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR1 = OFF      
// Table Read Protection bit (Block 1 (002000-003FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR2 = OFF      
// Table Read Protection bit (Block 2 (004000-005FFFh) not protected from table 
// reads executed in other blocks)
#pragma config EBTR3 = OFF      
// Table Read Protection bit (Block 3 (006000-007FFFh) not protected from table 
// reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      
// Boot Block Table Read Protection bit (Boot block (000000-0007FFh) 
// not protected from table reads executed in other blocks)

    /* -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+*/
// Initialise starts here:
	PORTA 	= 0x00; // Reset all Flip-Flops at PORTA 
	LATA	= 0x00;	
	/* All the PortA is set to be digital I/O. Remember that the PORTA 
     * can be also defined as analogue inputs*/
    ADCON1 	= 0x0F;	

    TRISA 	= 0b00011000;  	// PORTA A3, A2

	PORTB 	= 0x00; 		// Reset all Flip-Flops at PORTB  
	LATB	= 0x00;
	TRISB 	= 0b00000111;   // PORTB Cin, A1, A0, 
    
    PORTC 	= 0x00;         // Reset all Flip-Flops at PORTD   
	LATC	= 0x00;
	TRISC 	= 0b11110000;	// PORTC B3, B2, B1, B0

	PORTD 	= 0x00;         // Reset all Flip-Flops at PORTD   
	LATD	= 0x00;
	TRISD 	= 0b00000000;	// This port is left for the LCD interface

    GIE 	= 0;            // No interrupts in this very simple application
    
    
    Var_Cin_pre = 0;
    Var_A_pre = 0;
    Var_B_pre = 0;
    Var_Error_pre = 0;   
 
 
    // LCD initialisation 
	lcd_init(FOURBIT_MODE); 	/* Initialise the LCD module  */
	Var_LCD_flag = 1;   // Let's leave the LCD ready for printing the 1st result
    
}
  
/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables 
 many of the variables are "char" because we can easily monitor them using
 the watch window for debugging purposes.
  =========================================================================== */
void read_inputs (void){         

    // Reading Cin
// 1) Read all the bits (like LD a parallel 8-bit register)
    Value_data = PORTB;    
// 2) Mask the bit of interest (bitwise AND)    
    Value_data = Value_data & 0b00000100; 
// 3) Bitwise shift right 2 bits.     
    Var_Cin = Value_data >> 2; 
    
    // Reading A    
//   1) Read all the bits (like LD a parallel 8-bit register)
    Value_data2 = PORTA;    
//// 2) Mask the bit of interest (bitwise AND)    
    Value_data2 = Value_data2 & 0b00011000; 
    Value_data2 = Value_data2>>1; 
// 
     
    Value_data = PORTB;    
// 2) Mask the bit of interest (bitwise AND)  
    Value_data = Value_data & 0b00000011;     
    Var_A = Value_data | Value_data2;
//    
    
// Reading B
// 1) Read all the bits (like LD a parallel 8-bit register)
    Value_data = PORTC;    
// 2) Mask the bit of interest (bitwise AND) and shift
    Value_data = Value_data & 0b11110000; 
    Value_data = Value_data >>4;   
    Var_B = Value_data;
      
      
      // Detecting the error condition
      if (Var_B > 9 || Var_A > 9 )
          Var_Error = 1;
      else 
          Var_Error = 0;
      
      // Write if there is new information
    if (Var_Error != Var_Error_pre){  
        Var_LCD_flag = 1;
    }
    Var_Error_pre = Var_Error; 
    
    if ((Var_B != Var_B_pre)||(Var_A != Var_A_pre)||(Var_Cin != Var_Cin_pre) ){  
        Var_LCD_flag = 1; // Write only f new operands
    }
    
    Var_B_pre = Var_B; 
    Var_A_pre = Var_A; 
    Var_Cin_pre = Var_Cin;
 
}

/* =============================================================================
 The behavioural description of the circuit operation or the data processing
 algorithm.
 When this function is executed, all the variables are in the software domain,
 meaning that they are completely independent of the hardware, thus the
 algorithm is compatible among microcontroller families and pin connexions.
  =========================================================================== */
void truth_table(void){

    // Calculate the truth table only when new operands and no error
    if (Var_LCD_flag == 1) {    

        if (Var_Error == 0) {    //calculate the truth table    
            Value_data = Var_A + Var_B + Var_Cin; // Binary sum


                if (Value_data > 9)
                    Value_data = Value_data + 6; 		// BCD correction 
            //  else 
            //      Value_data = Value_data + 0; 

            Var_Cout = (Value_data & 0b00010000 )>> 4; 	// Cout
            Var_BCD_Units = Value_data & 0b00001111; 	      
        }
        else {
            Var_Cout = 0; 
            Var_BCD_Units = 0; 
        }
    
    }  

}   

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting 
	microcontroller variables into electrical signals.    
   ========================================================================== */
void write_outputs(void){
    
    if (Var_LCD_flag == 1) {    //write LCD only if new information
        
        if (Var_Error == 1) { 
            lcd_clear();		
            lcd_home();		
            lcd_puts("BCD CODE ERROR  ");
            lcd_home2();	
            lcd_puts("Input new A or B");
        } 
        else {  // Print results
        // Calculating the decimal equivalent to translate it to ASCII
            Var_Decimal = Var_Cout*10 + Var_BCD_Units;        
            sprintf(Var_Decimal_ASCII, "%u", Var_Decimal);
            lcd_clear();		
            lcd_home();		
            lcd_puts(".. Operating ...");
            lcd_home2();
            lcd_puts("BCD addition= ");
            lcd_goto2(14);
            lcd_puts(Var_Decimal_ASCII);
        }
    }                
    Var_LCD_flag = 0;   //Write only one time until new operand are read
}
